from functools import reduce
from pathlib import Path
import random
import re

import pytest

from betterpassphrase.config import BUFFER
from betterpassphrase.models import PartsOfSpeech
from betterpassphrase.generator import generate_phrase
from betterpassphrase.cli import main as betterpassphrase_cli
from betterpassphrase.mappings import (
    UNIT_PHRASE_LENGTHS,
    UNIT_PHRASE_MIN_LENGTH,
    UNIT_PHRASE_MAX_LENGTH,
)


extract_capitals = lambda text: re.findall(r"[A-Z][a-z]+", text)

def test_basic_phrase_generation():
    phrase = generate_phrase(length=6, sep=" ")
    assert isinstance(phrase.passphrase, str)
    assert len(phrase.passphrase.split(" ")) == 6
    assert phrase.word_count == 6


def test_basic_phrase_generation():
    phrase = generate_phrase(length=6, sep=" ", capitalize=True)
    assert isinstance(phrase.passphrase, str)
    assert len(extract_capitals(phrase.passphrase)) == 6
    assert phrase.word_count == 6


def test_phrase_with_separator():
    phrase = generate_phrase(length=4, sep="-")
    assert "-" in phrase.passphrase
    assert len(phrase.passphrase.split("-")) == 4


def test_word_count_and_combinations():
    in_length = random.choice(UNIT_PHRASE_LENGTHS)
    phrase = generate_phrase(length=in_length, sep=" ")
    assert phrase.word_count == len(phrase.words) == len(phrase.combination)

    out_length = UNIT_PHRASE_MAX_LENGTH + BUFFER
    phrase = generate_phrase(length=out_length, sep=" ")
    assert phrase.word_count == len(phrase.words) == len(phrase.combination)


def test_phrase_capitalization():
    phrase = generate_phrase(length=3, capitalize=True)
    assert all(word[0].isupper() for word in phrase.passphrase.split())

    phrase_lower = generate_phrase(length=3, capitalize=False)
    assert all(word[0].islower() for word in phrase_lower.passphrase.split())


def test_invalid_length():
    with pytest.raises(ValueError, match="Cannot generate phrase of length"):
        generate_phrase(
            length=-2
        )


def test_minimum_length():
    phrase = generate_phrase(length=UNIT_PHRASE_MIN_LENGTH, sep=" ")
    assert len(phrase.passphrase.split(" ")) == UNIT_PHRASE_MIN_LENGTH


def test_maximum_length():
    phrase = generate_phrase(length=UNIT_PHRASE_MAX_LENGTH, sep=" ")
    assert len(phrase.passphrase.split(" ")) == UNIT_PHRASE_MAX_LENGTH


def test_one_of_calculation():
    length = random.choice(UNIT_PHRASE_LENGTHS)
    phrase = generate_phrase(length=length)
    one_of_calculated = reduce(
        lambda x, y: x * y,
        [pos.n for pos in phrase.combination],
        1,
    )
    assert isinstance(phrase.one_of, int)
    assert phrase.one_of > 0  # `one_of` must be positive
    assert phrase.one_of == one_of_calculated


def test_parts_of_speech_loading():
    for part in PartsOfSpeech:
        words = part.words
        assert isinstance(words, list)
        assert all(isinstance(word, str) for word in words)
        assert len(words) > 0  # Ensure words are loaded


def test_random_word_selection():
    for part in PartsOfSpeech:
        word = part.word
        assert isinstance(word, str)
        assert word in part.words


def test_cli_integration():
    min_length = UNIT_PHRASE_MAX_LENGTH + BUFFER
    max_length = min_length + BUFFER
    temp_file = Path(".temp.txt")
    command_args_str = f"-l {min_length} -s - -c -o {temp_file}"
    betterpassphrase_cli(command_args_str.split(" "))
    result = temp_file.read_text()
    assert temp_file.exists()
    assert min_length <= len(result.strip().split("-")) <= max_length
    assert min_length <= len(extract_capitals(result)) <= max_length
    temp_file.unlink()



import secrets
from pathlib import Path


BUFFER = 3
"""
Buffer to add to the length of the passphrase.

This when added to length defines the maximum length of the passphrase.
"""

assert BUFFER >= 3, "Buffer must be at least 3 as the program might go into an infinite loop."

PARTS_OF_SPEECH_DIR = Path(__file__).parent / "parts_of_speech"
"""Path to the directory containing the parts of speech files."""

SEED: int | None = None
"""Seed for the random number generator."""

try:
    RANDOM_SELECTOR = secrets.SystemRandom(SEED).choice
except:
    RANDOM_SELECTOR = secrets.choice



from enum import Enum
from pathlib import Path
from typing import NamedTuple
from functools import cached_property, reduce

from .config import RANDOM_SELECTOR, PARTS_OF_SPEECH_DIR


class PartsOfSpeech(str, Enum):
    """
    Enum representing the different parts of speech used in the library.
    """
    ADJECTIVE = "adjectives.txt"
    SUBJECT_NOUN = "sub_nouns.txt"
    VERB = "verbs.txt"
    PREPOSITION = "prepositions.txt"
    DETERMINER = "determiners.txt"
    OBJECT_NOUN = "obj_nouns.txt"
    ADVERB = "adverbs.txt"
    CONJUNCTION = "conjunctions.txt"
    
    @property
    def filepath(self) -> Path:
        """
        The path to the file containing the words for the given part of speech.

        Returns:
            Path: The path to the file.
        """
        return PARTS_OF_SPEECH_DIR / self.value

    @cached_property
    def words(self) -> list[str]:
        """
        List of words for the given part of speech.

        Returns:
            list[str]: List of words.
        """
        wordfile = self.filepath
        if not wordfile.exists():
            return []
        return [
            x.strip()
            for x in wordfile.read_text().splitlines()
        ]
    
    @property
    def word(self) -> str:
        """
        A random word from the list of words for the given part of speech.

        Returns:
            str: A random word.
        """
        if not self.words:
            return ""
        return RANDOM_SELECTOR(self.words)
    
    @property
    def n(self) -> int:
        """
        The number of words in the list of words for the given part of speech.

        Returns:
            int: The number of words.
        """
        return len(self.words)

    def __str__(self) -> str:
        """
        The string representation of the part of speech.

        Returns:
            str: The string representation.
        """
        return self.value
    
    def __repr__(self) -> str:
        """
        The representation of the part of speech.

        Returns:
            str: The representation.
        """
        return self.value


class Passphrase(NamedTuple):
    """A generated passphrase with metadata."""

    words: list[str]
    """The list of words in the passphrase."""

    word_count: int
    """The number of words in the passphrase."""

    separator: str
    """The separator between words in the passphrase."""

    capitalize: bool
    """Whether the words in the passphrase should be capitalized."""

    sub_combinations: list[list[PartsOfSpeech]]
    """A list of sub-combinations of parts of speech used to generate the passphrase."""

    @property
    def passphrase(self) -> str:
        """The generated passphrase."""
        return self.separator.join(self.words)

    @property
    def combination(self) -> list[PartsOfSpeech]:
        """The combination of parts of speech used to generate the passphrase."""
        return [
            pos for phrase in self.sub_combinations for pos in phrase
        ]

    @property
    def one_of(self) -> int:
        """This is one of the different passphrases can be generated using the same set of parts of speech wordlists."""
        return reduce(lambda x, y: x * y, [pos.n for pos in self.combination])

    @property
    def wordlist_probability(self) -> float:
        """The probability of generating the passphrase using the same set of parts of speech wordlists."""
        return 1 / self.one_of
    
    @property
    def character_one_of(self) -> int:
        """This is one of the different passphrases can be generated using the same of alphabetical characters."""
        return reduce(
            lambda x, y: x * y,
            [
                26 if char.isalpha() else 10 if char.isdigit() else 1
                for char in "".join(self.words)
            ],
        )

    @property
    def character_probability(self) -> float:
        """The probability of generating the passphrase using alphabetical characters."""
        return 1 / self.character_one_of

    def __str__(self) -> str:
        """The string representation of the passphrase."""
        return self.passphrase


P = PartsOfSpeech
"""
A short alias for the PartsOfSpeech class.
"""



"""
The main module for the betterpassphrase package.

Contains the following submodules:
- `generator`: Contains the `generate_phrase` function and the `Passphrase` class.
- `config`: Contains the `PARTS_OF_SPEECH_DIR` constant, which is the path to the directory containing the parts of speech files.
"""
from .generator import generate_phrase, Passphrase
from .config import PARTS_OF_SPEECH_DIR

__all__ = ["generate_phrase", "Passphrase", "PARTS_OF_SPEECH_DIR"]



from .models import P


LENGTH_TO_WORD_COMBINATIONS_MAP = {
    3: [
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN],      # The Quick Hunter
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN],       # The Quick Wolf
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB],           # The Hunter Chased
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB],            # The Wolf Chased
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.ADVERB],          # Quick Hunter Eagerly
        [P.ADJECTIVE, P.OBJECT_NOUN, P.ADVERB],           # Quick Wolf Eagerly
        [P.SUBJECT_NOUN, P.VERB, P.ADVERB],               # Hunter Chased Eagerly
        [P.OBJECT_NOUN, P.VERB, P.ADVERB],                # Wolf Chased Eagerly
        [P.VERB, P.DETERMINER, P.OBJECT_NOUN],            # Chased The Wolf
        [P.VERB, P.DETERMINER, P.SUBJECT_NOUN],           # Chased The Hunter
        [P.ADVERB, P.PREPOSITION, P.DETERMINER],          # Eagerly Across The
        [P.PREPOSITION, P.DETERMINER, P.ADJECTIVE],       # Across The Quick
        [P.PREPOSITION, P.DETERMINER, P.SUBJECT_NOUN],    # Across The Hunter
    ],
    4: [
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB],             # The Quick Hunter Chased
        [P.DETERMINER, P.OBJECT_NOUN, P.ADVERB, P.VERB],                 # The Wolf Eagerly Chased
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.ADVERB],                 # Quick Hunter Chased Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.PREPOSITION],           # The Hunter Chased Across
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.ADVERB],                 # The Wolf Chased Eagerly
        [P.SUBJECT_NOUN, P.OBJECT_NOUN, P.VERB, P.ADVERB],               # Hunter Wolf Chased Eagerly
        [P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN],           # Hunter Chased The Wolf
        [P.DETERMINER, P.OBJECT_NOUN, P.CONJUNCTION, P.SUBJECT_NOUN],    # The Wolf And Hunter
        [P.OBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.SUBJECT_NOUN],    # Wolf And The Hunter
        [P.OBJECT_NOUN, P.ADVERB, P.VERB, P.SUBJECT_NOUN],               # Wolf Eagerly Chased Hunter
        [P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN],    # Hunter And The Wolf
        [P.DETERMINER, P.SUBJECT_NOUN, P.CONJUNCTION, P.OBJECT_NOUN],    # The Hunter And Wolf
        [P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.SUBJECT_NOUN],           # Wolf Chased The Hunter
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB],              # The Quick Wolf Chased
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.OBJECT_NOUN],            # Quick Hunter Chased Wolf
        [P.DETERMINER, P.SUBJECT_NOUN, P.ADVERB, P.VERB],                # The Hunter Eagerly Chased
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.SUBJECT_NOUN],           # The Wolf Chased Hunter
        [P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION],                # Wolf Chased Eagerly Across
    ],
    5: [
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.ADVERB],                 # The Quick Hunter Chased Eagerly
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.ADJECTIVE, P.SUBJECT_NOUN],            # The Wolf Chased Quick Hunter
        [P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN],            # Hunter Chased The Quick Wolf
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.ADVERB],                  # The Quick Wolf Chased Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN],           # The Hunter Chased The Wolf
        [P.DETERMINER, P.OBJECT_NOUN, P.ADVERB, P.VERB, P.SUBJECT_NOUN],               # The Wolf Eagerly Chased Hunter
        [P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],               # Hunter Chased The Wolf Eagerly
        [P.SUBJECT_NOUN, P.VERB, P.ADJECTIVE, P.OBJECT_NOUN, P.PREPOSITION],           # Hunter Chased Quick Wolf Across
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.CONJUNCTION, P.OBJECT_NOUN],     # The Quick Hunter And Wolf
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN],     # Quick Hunter And The Wolf
        [P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN],            # Wolf Chased The Quick Hunter
        [P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.SUBJECT_NOUN],            # Quick Wolf Chased The Hunter
        [P.DETERMINER, P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN],    # The Hunter And The Wolf
        [P.ADVERB, P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.PREPOSITION],                # Eagerly The Wolf Chased Across
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN],            # Quick Hunter Chased The Wolf
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION],                # The Wolf Chased Eagerly Across
        [P.SUBJECT_NOUN, P.VERB, P.ADJECTIVE, P.OBJECT_NOUN, P.ADVERB],                # Hunter Chased Quick Wolf Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.CONJUNCTION, P.ADJECTIVE, P.OBJECT_NOUN],     # The Hunter And Quick Wolf
    ],
    6: [
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN],           # The Quick Hunter Chased The Wolf
        [P.DETERMINER, P.SUBJECT_NOUN, P.ADVERB, P.VERB, P.DETERMINER, P.OBJECT_NOUN],              # The Hunter Eagerly Chased The Wolf
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],               # Quick Hunter Chased The Wolf Eagerly
        [P.DETERMINER, P.OBJECT_NOUN, P.ADVERB, P.VERB, P.DETERMINER, P.SUBJECT_NOUN],              # The Wolf Eagerly Chased The Hunter
        [P.ADVERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB],               # Eagerly Across The Quick Hunter Chased
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],              # The Hunter Chased The Wolf Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION],                # The Quick Wolf Chased Eagerly Across
        [P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.ADVERB],             # Hunter And The Wolf Chased Eagerly
        [P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.ADVERB],               # Wolf Chased The Quick Hunter Eagerly
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN, P.VERB],          # Quick Hunter And The Wolf Chased
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.OBJECT_NOUN],    # The Quick Hunter And The Wolf
        [P.ADVERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB],                # Eagerly Across The Quick Wolf Chased
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN],           # The Wolf Chased The Quick Hunter
        [P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION, P.DETERMINER, P.SUBJECT_NOUN],             # Wolf Chased Eagerly Across The Hunter
        [P.DETERMINER, P.SUBJECT_NOUN, P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION],             # The Hunter Wolf Chased Eagerly Across
    ],
    7: [
        [P.DETERMINER, P.OBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.ADVERB],       # The Wolf And The Hunter Chased Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.OBJECT_NOUN, P.VERB, P.ADJECTIVE, P.CONJUNCTION, P.ADVERB],        # The Hunter Wolf Chased Quick And Eagerly
        [P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN, P.ADJECTIVE, P.CONJUNCTION, P.ADVERB],        # Hunter Chased The Wolf Quick And Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.SUBJECT_NOUN, P.ADVERB],         # The Quick Wolf Chased The Hunter Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.ADVERB],         # The Hunter Chased The Quick Wolf Eagerly
        [P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],        # Quick Hunter Chased Across The Wolf Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.CONJUNCTION, P.OBJECT_NOUN, P.VERB, P.PREPOSITION, P.ADVERB],      # The Hunter And Wolf Chased Across Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.PREPOSITION, P.DETERMINER, P.OBJECT_NOUN, P.VERB],    # The Quick Hunter Across The Wolf Chased
        [P.SUBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN],        # Hunter Chased Eagerly Across The Quick Wolf
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.ADVERB],         # The Wolf Chased The Quick Hunter Eagerly
        [P.OBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.ADVERB],        # Wolf Chased Across The Quick Hunter Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],         # The Quick Hunter Chased The Wolf Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.SUBJECT_NOUN, P.VERB],    # The Quick Wolf And The Hunter Chased
    ],
    8: [
        [P.DETERMINER, P.OBJECT_NOUN, P.ADVERB, P.VERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN],     # The Wolf Eagerly Chased Across The Quick Hunter
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.ADVERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN],     # The Wolf Chased Eagerly Across The Quick Hunter
        [P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.OBJECT_NOUN, P.ADVERB],     # The Quick Hunter Chased Across The Wolf Eagerly
        [P.DETERMINER, P.OBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.SUBJECT_NOUN, P.ADVERB],     # The Wolf Chased Across The Quick Hunter Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.ADVERB],     # The Hunter Chased Across The Quick Wolf Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.OBJECT_NOUN, P.VERB, P.PREPOSITION, P.ADJECTIVE, P.CONJUNCTION, P.ADVERB],    # The Hunter Wolf Chased Across Quick And Eagerly
        [P.DETERMINER, P.SUBJECT_NOUN, P.CONJUNCTION, P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.ADVERB],     # The Hunter And The Quick Wolf Chased Eagerly
        [P.DETERMINER, P.ADJECTIVE, P.OBJECT_NOUN, P.VERB, P.PREPOSITION, P.DETERMINER, P.SUBJECT_NOUN, P.ADVERB],     # The Quick Wolf Chased Across The Hunter Eagerly
    ],
}
"""
Map of lengths to word combinations.
Here's a better representation:
```yaml
LENGTH_TO_WORD_COMBINATIONS_MAP:
    length1: [
        [option1 for that length1],
        [option2 for that length1],
        ...
    ]
    length2: [
        [option1 for that length2],
        [option2 for that length2],
        ...
    ]
    ...
```
"""

UNIT_PHRASE_LENGTHS = sorted(LENGTH_TO_WORD_COMBINATIONS_MAP)
"""List of lengths that can be used to generate an unit passphrase."""

UNIT_PHRASE_MIN_LENGTH = UNIT_PHRASE_LENGTHS[0]
"""Minimum length allowed to generate an unit passphrase."""

UNIT_PHRASE_MAX_LENGTH = UNIT_PHRASE_LENGTHS[-1]
"""Maximum length allowed to generate an unit passphrase."""



import logging
import traceback
import concurrent.futures
from typing import Callable, Iterable


def get_trace(e: Exception, n: int = 5):
    """Get the last n lines of the traceback for an exception"""
    return "".join(traceback.format_exception(e)[-n:])

def run_parallel_exec(exec_func: Callable, iterable: Iterable, *func_args, **kwargs):
    """
    Runs the `exec_func` function in parallel for each element in the `iterable` using a thread pool executor.
    
    Parameters:
        exec_func (Callable): The function to be executed for each element in the `iterable`.
        iterable (Iterable): The collection of elements for which the `exec_func` function will be executed.
        *func_args: Additional positional arguments to be passed to the `exec_func` function.
        **kwargs: Additional keyword arguments to customize the behavior of the function.
            - max_workers (int): The maximum number of worker threads in the thread pool executor. Default is 100.
            - quiet (bool): If True, suppresses the traceback logging for exceptions. Default is False.
    
    Returns:
        list[tuple]: A list of tuples where each tuple contains the element from the `iterable` and the result of executing the `exec_func` function on that element.

    Example:
        >>> from app.utils.helpers import run_parallel_exec
        >>> run_parallel_exec(lambda x: str(x), [1, 2, 3])
        [(1, '1'), (2, '2'), (3, '3')]
    """
    func_name = f"{exec_func.__name__} | parallel_exec | " if hasattr(exec_func, "__name__") else "unknown | parallel_exec | "
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=kwargs.pop("max_workers", 100), thread_name_prefix=func_name
    ) as executor:
        # Start the load operations and mark each future with each element
        future_element_map = {
            executor.submit(exec_func, element, *func_args): element
            for element in iterable
        }
        result: list[tuple] = []
        for future in concurrent.futures.as_completed(future_element_map):
            element = future_element_map[future]
            try:
                result.append((element, future.result()))
            except Exception as exc:
                log_trace = exc if kwargs.pop("quiet", False) else get_trace(exc, 3)
                logging.error(f"Got error while running parallel_exec: {element}: \n{log_trace}")
                result.append((element, exc))
        return result

def run_parallel_exec_but_return_in_order(exec_func: Callable, iterable: Iterable, *func_args, **kwargs):
    """
    Runs the `exec_func` function in parallel for each element in the `iterable` using a thread pool executor.
    Returns the result in the same order as the `iterable`.
    """
    # note this is usable only when iterable has types that are hashable
    result = run_parallel_exec(exec_func, iterable:=list(iterable), *func_args, **kwargs)

    # sort the result in the same order as the iterable
    result.sort(key=lambda x: iterable.index(x[0]))

    return [x[-1] for x in result]



from functools import reduce

from .config import BUFFER, RANDOM_SELECTOR
from .models import P, Passphrase
from .mappings import (
    UNIT_PHRASE_LENGTHS,
    UNIT_PHRASE_MAX_LENGTH,
    LENGTH_TO_WORD_COMBINATIONS_MAP,
)


def generate_lengths(length: int = 10, buffer: int = 3) -> list[int]:
    """
    Generate a list of word lengths for a passphrase generator.

    This function creates a list of word lengths that adhere to the constraints of
    a desired total passphrase length (`length`) and an allowable buffer (`buffer`).
    Word lengths are chosen from a predefined list (`UNIT_PHRASE_LENGTHS`), with
    separators of length 1 (representing conjunctions) included between words.

    The algorithm ensures the following:
    - The total length of the passphrase (including separators) is between
      `length` and `length + buffer`.
    - Word lengths of 1 are not used, except for separators.

    Args:
        length (int): The target length of the passphrase.
        buffer (int): The maximum additional length allowed beyond the target length.
                      Must be at least 3.

    Returns:
        list[int]: A list of integers representing word lengths for the passphrase.
                   Separators are not explicitly included in the output.

    Notes:
        - The function ensures that separators of length 1 are used during generation
          but are removed from the final output list.
        - The function relies on a predefined `UNIT_PHRASE_LENGTHS` list, which should
          include acceptable word lengths.

    Examples:
        >>> UNIT_PHRASE_LENGTHS = [3, 4, 5, 6, 7, 8]
        >>> generate_lengths(length=10, buffer=3)
        [5, 4]  # Example output, actual values may vary due to randomness.
    """
    lengths = []
    # Ensure a minimum buffer value of 3
    buffer = max(3, buffer)

    # UNIT_PHRASE_LENGTHS should be a predefined list of acceptable word lengths.
    options = UNIT_PHRASE_LENGTHS
    if 1 in options:

        # Remove 1 as a valid word length (used only for separators).
        options.remove(1)

    while sum(lengths[:-1]) < length:
        # Calculate the difference between the target length and current sum of lengths
        diff = length - sum(lengths[:-1]) - 1  # -1 for the added separator

        if diff in options:
            # If the difference matches an option, use it as the next word length
            sub_length = diff
        else:
            # Otherwise, choose a random word length from the available options
            sub_length = RANDOM_SELECTOR(options)

        lengths.append(sub_length)

        # Check if the total length exceeds the allowed range
        if sum(lengths) > length + buffer:
            # If the length exceeds, remove the last added word and try again
            lengths = lengths[:-1]
            continue

        # Add a separator (length of 1) after the word length
        lengths.append(1)

    # Remove any standalone separators (lengths of 1) from the final list
    lengths = [x for x in lengths if not x == 1]

    return lengths


def generate_phrase(
    length: int = 6, sep: str = "", capitalize: bool = True
) -> Passphrase:
    """
    Generate a passphrase of the specified length.

    Args:
        length (int): The number of words in the passphrase.
        sep (str): Separator between words.
        capitalize (bool): Whether to capitalize words.

    Returns:
        Passphrase: Generated passphrase with metadata.
    """
    # If length is greater than the maximum length, generate a phrase of the maximum length
    # and then generate a new phrase of the remaining length, joining them with the conjunction.
    if length > UNIT_PHRASE_MAX_LENGTH:
        words: list[str] = []
        sub_combinations: list[list[P]] = []

        # Generate a list of sub-phrase lengths
        lengths = generate_lengths(length, buffer=BUFFER)
        total_length = (
            sum(lengths)        # Total number of words in the sub-phrases
            + len(lengths) - 1  # Number of conjunctions
        )

        for index, curr_length in enumerate(lengths):
            # Generate a sub-phrase of the current length and add it to the list
            passphrase = generate_phrase(curr_length, sep, capitalize)
            words.extend(passphrase.words)
            sub_combinations.append(passphrase.combination)

            if index == len(lengths) - 1:
                continue

            # Get the conjunction word and capitalize it if necessary
            conjunction = (
                P.CONJUNCTION.word.capitalize() if capitalize else P.CONJUNCTION.word
            )
            words.append(conjunction)
            sub_combinations.append([P.CONJUNCTION])

        # Join the generated phrases with the separator and return the result
        return Passphrase(
            words=words, 
            word_count=total_length, 
            separator=sep,
            capitalize=capitalize,
            sub_combinations=sub_combinations,
        )

    # If the length is not in the length-to-word-combinations map, raise a ValueError
    if length not in LENGTH_TO_WORD_COMBINATIONS_MAP:
        raise ValueError(f"Cannot generate phrase of length {length}")

    # Choose a random combination of parts of speech for the given length
    combination = RANDOM_SELECTOR(LENGTH_TO_WORD_COMBINATIONS_MAP[length])

    # NOTE: Uncomment this for debugging (will print the selected combination and its index)
    # print(
    #     f"COMBINATION LENGTH:INDEX:[*PARTS] : {length}"
    #     f":{LENGTH_TO_WORD_COMBINATIONS_MAP[length].index(combination)}"
    #     f":[{', '.join([x.name for x in combination])}]"
    # )

    # Generate the words for the combination and capitalize them if necessary
    words = [pos.word for pos in combination]
    if capitalize:
        words = [word.capitalize() for word in words]

    # Return the generated passphrase with its metadata
    return Passphrase(
        words=words, 
        word_count=length, 
        separator=sep,
        capitalize=capitalize,
        sub_combinations=[combination],
    )



import argparse

from betterpassphrase.models import Passphrase
from betterpassphrase.utils import run_parallel_exec_but_return_in_order
from .generator import generate_phrase


def main(_args: list[str] = None):
    parser = argparse.ArgumentParser(
        description="Generate a phrase based on the specified options."
    )

    # -l flag for phrase length
    parser.add_argument(
        "-l",
        "--length",
        type=int,
        default=6,
        help="Number of words in the phrase (default: 6).",
    )

    # -s flag for separator
    parser.add_argument(
        "-s",
        "--sep",
        type=str,
        default="-",
        help="Separator between words (default: no separator).",
    )

    # -c flag for capitalization (True by default)
    parser.add_argument(
        "-c",
        "--capitalize",
        action="store_true",
        help="Capitalize the phrase (default: True).",
    )

    # -o flag for output file
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="",
        help="File to write the phrase to (default: print to stdout).",
    )
    
    # -r flag for generating multiple phrases
    parser.add_argument(
        "-n",
        "--num-phrases",
        type=int,
        default=1,
        help="Number of phrases to generate (default: 1).",
    )

    # -v flag for verbosity level
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        choices=[0, 1, 2],
        default=0,
        help="Verbosity level: 0 for passphrase only, 1 for basic info, 2 for detailed info (default: 1).",
    )

    args = parser.parse_args(_args)

    capitalize = args.capitalize

    if args.num_phrases >= 1:
        phrases: list[Passphrase] = run_parallel_exec_but_return_in_order(
            generate_phrase,
            [args.length] * args.num_phrases, 
            args.sep, 
            capitalize, 
        )
    else:
        print("Invalid number of phrases to generate.")
        exit(1)

    for i, phrase in enumerate(phrases):
        if args.output:
            _mode = "w" if i == 0 else "a"
            with open(args.output, _mode) as f:
                f.write(f"{phrase.passphrase}\n")

        pos_sep = ", "
        phrase_sep = "\n                        "

        if args.verbosity == 0:
            # Only the passphrase
            print(phrase.passphrase)
        elif args.verbosity == 1:
            # Basic info
            print()
            print(f"Generated phrase: {phrase.passphrase}")
            print(f"Word count:       {phrase.word_count}")
        elif args.verbosity == 2:
            # Detailed info
            print()
            print(f"Generated phrase:       {phrase.passphrase}")
            print(f"Word count:             {phrase.word_count}")
            print(f"Wordlist Probability:   {phrase.wordlist_probability:.2e}")
            print(f"Character Probability:  {phrase.character_probability:.2e}")
            print(f"Parts of speech:        {phrase_sep.join(pos_sep.join(pos.name.lower() for pos in phrase) for phrase in phrase.sub_combinations)}\n")


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
    name="BetterPassphrase",
    version="0.1.8",
    description="A Python library to generate secure, meaningful passphrases.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Subhayu Kumar Bala",
    author_email="balasubhayu99@gmail.com",
    url="https://github.com/subhayu99/BetterPassphrase",
    packages=find_packages(),
    license="MIT",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "betterpassphrase=betterpassphrase.cli:main",
        ],
    },
)
